"""Define API endpoints for nodes."""
from dataclasses import dataclass
from datetime import datetime
from typing import Awaitable, Callable, List

from defusedxml import ElementTree
import maya


@dataclass
class Note:  # pylint: disable=too-many-instance-attributes
    """Define a representation of a Pinboard note."""

    note_id: str
    title: str
    hash: str
    created_at: datetime
    updated_at: datetime
    length: int


def async_create_note_from_xml(tree: ElementTree) -> Note:
    """Create a note from an XML response.

    :param tree: The XML element tree to parse
    :type tree: ``ElementTree``
    :rtype: ``Note``
    """
    return Note(
        tree.attrib["id"],
        tree[0].text,
        tree[1].text,
        maya.parse(tree[2].text).datetime(),
        maya.parse(tree[3].text).datetime(),
        int(tree[4].text),
    )


class NoteAPI:  # pylint: disable=too-few-public-methods
    """Define a note "manager" object."""

    def __init__(self, async_request: Callable[..., Awaitable]) -> None:
        """Initialize."""
        self._async_request = async_request

    async def async_get_notes(self) -> List[Note]:
        """Get all notes.

        :rtype: ``List[Note]``
        """
        resp = await self._async_request("get", "notes/list")
        return [async_create_note_from_xml(note) for note in resp]



"""Define API endpoints for bookmarks."""
from dataclasses import dataclass
from datetime import date, datetime
from typing import Any, Awaitable, Callable, Dict, List, Optional

from defusedxml import ElementTree
import maya

DEFAULT_RECENT_BOOKMARKS_COUNT: int = 15


@dataclass
class Bookmark:  # pylint: disable=too-many-instance-attributes
    """Define a representation of a Pinboard bookmark."""

    hash: str
    href: str
    title: str
    description: str
    last_modified: datetime
    tags: List[str]
    unread: bool
    shared: bool


def async_create_bookmark_from_xml(tree: ElementTree) -> Bookmark:
    """Create a bookmark from an XML response.

    :param tree: The XML element tree to parse
    :type tree: ``ElementTree``
    :rtype: ``Bookmark``
    """
    return Bookmark(
        tree.attrib["hash"],
        tree.attrib["href"],
        tree.attrib["description"],
        tree.attrib["extended"],
        maya.parse(tree.attrib["time"]).datetime(),
        tree.attrib["tag"].split(),
        tree.attrib.get("toread") == "yes",
        tree.attrib.get("shared") == "yes",
    )


class BookmarkAPI:
    """Define an API "manager" object."""

    def __init__(self, async_request: Callable[..., Awaitable]) -> None:
        """Initialize."""
        self._async_request = async_request

    async def async_add_bookmark(
        self,
        url: str,
        title: str,
        *,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        created_datetime: Optional[datetime] = None,
        replace: bool = True,
        shared: bool = False,
        toread: bool = False,
    ) -> None:
        """Add a new bookmark.

        :param url: The URL of the bookmark
        :type url: ``str``
        :param title: The title of the bookmark
        :type title: ``str``
        :param description: The optional description of the bookmark
        :type description: ``Optional[str]``
        :param tags: An optional list of tags to assign to the bookmark
        :type tags: ``Optional[List[str]]``
        :param created_datetime: The optional creation datetime to use (defaults to now)
        :type created_datetime: ``Optional[datetime]``
        :param replace: Whether this should replace a bookmark with the same URL
        :type replace: ``bool``
        :param shared: Whether this bookmark should be shared
        :type shared: ``bool``
        :param toread: Whether this bookmark should be unread
        :type toread: ``bool``
        :rtype: ``Bookmark``
        """
        params: Dict[str, Any] = {"url": url, "description": title}

        if description:
            params["description"] = description
        if tags:
            params["tags"] = tags
        if created_datetime:
            params["dt"] = created_datetime.isoformat()

        params["replace"] = "yes" if replace else "no"
        params["shared"] = "yes" if shared else "no"
        params["toread"] = "yes" if toread else "no"

        await self._async_request("get", "posts/add", params=params)

    async def async_delete_bookmark(self, url: str) -> None:
        """Delete a bookmark by URL.

        :param url: The URL of the bookmark to delete
        :type url: ``str``
        """
        await self._async_request("get", "posts/delete", params={"url": url})

    async def async_get_all_bookmarks(
        self,
        *,
        tags: Optional[List[str]] = None,
        start: int = 0,
        results: Optional[int] = None,
        from_dt: Optional[datetime] = None,
        to_dt: Optional[datetime] = None,
    ) -> List[Bookmark]:
        """Get recent bookmarks.

        :param tags: An optional list of tags to filter results by
        :type tags: ``Optional[List[str]]``
        :param start: The optional starting index to return (defaults to the start)
        :type start: ``int``
        :param results: The optional number of results (defaults to all)
        :type results: ``int``
        :param from_dt: The optional datetime to start from
        :type from_dt: ``Optional[datetime]``
        :param to_dt: The optional datetime to end at
        :type to_dt: ``Optional[datetime]``
        :rtype: ``List[Bookmark]``
        """
        params: Dict[str, Any] = {"start": start}

        if tags:
            params["tags"] = " ".join([str(tag) for tag in tags])
        if results:
            params["results"] = results
        if from_dt:
            params["fromdt"] = from_dt.isoformat()
        if to_dt:
            params["fromdt"] = to_dt.isoformat()

        resp = await self._async_request("get", "posts/all", params=params)
        return [async_create_bookmark_from_xml(bookmark) for bookmark in resp]

    async def async_get_bookmark_by_url(self, url: str) -> Optional[Bookmark]:
        """Get bookmark by a URL. Returns None if no bookmark exists for URL.

        :param url: The URL of the bookmark to get
        :type url: ``Optional[Bookmark]``
        :rtype: ``Optional[Bookmark]``
        """
        resp = await self._async_request("get", "posts/get", params={"url": url})

        try:
            return async_create_bookmark_from_xml(resp[0])
        except IndexError:
            return None

    async def async_get_bookmarks_by_date(
        self, bookmarked_on: date, *, tags: Optional[List[str]] = None
    ) -> List[Bookmark]:
        """Get bookmarks that were created on a specific date.

        :param bookmarked_on: The date to examine
        :type bookmarked_on: ``date``
        :param tags: An optional list of tags to filter results by
        :type tags: ``Optional[List[str]]``
        :rtype: ``List[Bookmark]``
        """
        params: Dict[str, Any] = {"dt": str(bookmarked_on)}

        if tags:
            params["tags"] = " ".join([str(tag) for tag in tags])

        resp = await self._async_request("get", "posts/get", params=params)
        return [async_create_bookmark_from_xml(bookmark) for bookmark in resp]

    async def async_get_dates(
        self, *, tags: Optional[List[str]] = None
    ) -> Dict[date, int]:
        """Get a dictionary of dates and the number of bookmarks created on that date.

        :param tags: An optional list of tags to filter results by
        :type tags: ``Optional[List[str]]``
        :rtype: ``Dict[date, int]``
        """
        params: Dict[str, Any] = {}

        if tags:
            params["tags"] = " ".join([str(tag) for tag in tags])

        resp = await self._async_request("get", "posts/dates")

        return {
            maya.parse(row.attrib["date"]).datetime().date(): int(row.attrib["count"])
            for row in resp
        }

    async def async_get_last_change_datetime(self) -> datetime:
        """Return the most recent time a bookmark was added, updated or deleted.

        :rtype: ``datetime``
        """
        resp = await self._async_request("get", "posts/update")
        maya_dt = maya.parse(resp.attrib["time"])
        return maya_dt.datetime()

    async def async_get_recent_bookmarks(
        self,
        *,
        count: int = DEFAULT_RECENT_BOOKMARKS_COUNT,
        tags: Optional[List[str]] = None,
    ) -> List[Bookmark]:
        """Get recent bookmarks.

        :param count: The number of bookmarks to return (max of 100)
        :type count: ``int``
        :param tags: An optional list of tags to filter results by
        :type tags: ``Optional[List[str]]``
        :rtype: ``List[Bookmark]``
        """
        params: Dict[str, Any] = {"count": count}

        if tags:
            params["tags"] = " ".join([str(tag) for tag in tags])

        resp = await self._async_request("get", "posts/recent", params=params)
        return [async_create_bookmark_from_xml(bookmark) for bookmark in resp]

    async def async_get_suggested_tags(self, url: str) -> Dict[str, List[str]]:
        """Return a dictionary of popular and recommended tags for a URL.

        :param url: The URL of the bookmark to delete
        :type url: ``str``
        """
        data: Dict[str, List[str]] = {"popular": [], "recommended": []}

        resp = await self._async_request("get", "posts/suggest", params={"url": url})
        for tag in resp:
            if tag.text not in data[tag.tag]:
                data[tag.tag].append(tag.text)

        return data



"""Define exception types for ``aiopinboard``."""
from defusedxml import ElementTree


class PinboardError(Exception):
    """Define a base Pinboard exception."""

    pass


class RequestError(PinboardError):
    """Define a exception related to HTTP request errors."""

    pass


def raise_on_response_error(response_root: ElementTree) -> None:
    """Raise an error if the data indicates that something went wrong."""
    if "code" not in response_root.attrib:
        return

    if response_root.attrib["code"] == "done":
        return

    raise RequestError(response_root.attrib["code"])



"""Define an API object to interact with the Pinboard API."""
import logging
from typing import Optional

from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError
from defusedxml import ElementTree

from aiopinboard.bookmark import BookmarkAPI
from aiopinboard.errors import RequestError, raise_on_response_error
from aiopinboard.note import NoteAPI
from aiopinboard.tag import TagAPI

_LOGGER = logging.getLogger(__name__)

API_URL_BASE: str = "https://api.pinboard.in/v1"

DEFAULT_TIMEOUT: int = 10


class API:  # pylint: disable=too-few-public-methods
    """Define an API object.

    :param api_token: A Pinboard API token
    :type api_token: ``str``
    :param session: An optional ``aiohttp`` ``ClientSession``
    :type api_token: ``Optional[ClientSession]``
    """

    def __init__(
        self, api_token: str, *, session: Optional[ClientSession] = None
    ) -> None:
        """Initialize."""
        self._api_token = api_token
        self._session: ClientSession = session

        self.bookmark = BookmarkAPI(self._async_request)
        self.note = NoteAPI(self._async_request)
        self.tag = TagAPI(self._async_request)

    async def _async_request(self, method: str, endpoint: str, **kwargs) -> ElementTree:
        """Make a request to the API and return the XML response."""
        kwargs.setdefault("params", {})
        kwargs["params"]["auth_token"] = self._api_token

        use_running_session = self._session and not self._session.closed

        if use_running_session:
            session = self._session
        else:
            session = ClientSession(timeout=ClientTimeout(total=DEFAULT_TIMEOUT))

        try:
            async with session.request(
                method, f"{API_URL_BASE}/{endpoint}", **kwargs
            ) as resp:
                resp.raise_for_status()
                body = await resp.text()

                _LOGGER.debug("Response text for %s: %s", endpoint, body)

                response_root = ElementTree.fromstring(body.encode("utf-8"))
                raise_on_response_error(response_root)

                return response_root
        except ClientError as err:
            raise RequestError(err) from None
        finally:
            if not use_running_session:
                await session.close()



"""Define API endpoints for tags."""
from typing import Awaitable, Callable, Dict


class TagAPI:
    """Define a tag "manager" object."""

    def __init__(self, async_request: Callable[..., Awaitable]) -> None:
        """Initialize."""
        self._async_request = async_request

    async def async_delete_tag(self, tag: str) -> None:
        """Delete a tag.

        :param tag: The tag to delete
        :type tag: ``str``
        """
        await self._async_request("get", "tags/delete", params={"tag": tag})

    async def async_get_tags(self) -> Dict[str, int]:
        """Get a mapping of all tags in this account and how many times each is used.

        :rtype: ``Dict[str, int]``
        """
        resp = await self._async_request("get", "tags/get")
        return {tag.attrib["tag"]: int(tag.attrib["count"]) for tag in resp}

    async def async_rename_tag(self, old: str, new: str) -> None:
        """Rename a tag.

        :param old: The tag to rename
        :type old: ``str``
        :param new: The new name
        :type new: ``str``
        """
        await self._async_request("get", "tags/rename", params={"old": old, "new": new})



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiopinboard']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.6.2,<4.0.0',
 'async_timeout>=3.0.1,<4.0.0',
 'defusedxml>=0.6.0,<0.7.0',
 'maya>=0.6.1,<0.7.0']

setup_kwargs = {
    'name': 'aiopinboard',
    'version': '0.1.2',
    'description': 'A Python 3, asyncio-based library for the Pinboard API',
    'long_description': '# 📌 aiopinboard: A Python 3 Library for Pinboard\n\n[![CI](https://github.com/bachya/aiopinboard/workflows/CI/badge.svg)](https://github.com/bachya/aiopinboard/actions)\n[![PyPi](https://img.shields.io/pypi/v/aiopinboard.svg)](https://pypi.python.org/pypi/aiopinboard)\n[![Version](https://img.shields.io/pypi/pyversions/aiopinboard.svg)](https://pypi.python.org/pypi/aiopinboard)\n[![License](https://img.shields.io/pypi/l/aiopinboard.svg)](https://github.com/bachya/aiopinboard/blob/master/LICENSE)\n[![Code Coverage](https://codecov.io/gh/bachya/aiopinboard/branch/dev/graph/badge.svg)](https://codecov.io/gh/bachya/aiopinboard)\n[![Maintainability](https://api.codeclimate.com/v1/badges/a03c9e96f19a3dc37f98/maintainability)](https://codeclimate.com/github/bachya/aiopinboard/maintainability)\n[![Say Thanks](https://img.shields.io/badge/SayThanks-!-1EAEDB.svg)](https://saythanks.io/to/bachya)\n\n`aiopinboard` is a Python3, `asyncio`-focused library for interacting with the\n[Pinboard](https://pinboard.in) API.\n\n- [Installation](#installation)\n- [Python Versions](#python-versions)\n- [API Token](#api-token)\n- [Usage](#usage)\n  * [Bookmarks](#bookmarks)\n    + [The `Bookmark` Object](#the--bookmark--object)\n    + [Getting the Last Change Datetime](#getting-the-last-change-datetime)\n    + [Getting Bookmarks](#getting-bookmarks)\n    + [Adding a Bookmark](#adding-a-bookmark)\n    + [Deleting a Bookmark](#deleting-a-bookmark)\n  * [Tags](#tags)\n    + [Getting Tags](#getting-tags)\n    + [Getting Suggested Tags](#getting-suggested-tags)\n    + [Deleting a Tag](#deleting-a-tag)\n    + [Renaming a Tag](#renaming-a-tag)\n  * [Notes](#notes)\n    + [The `Note` Object](#the--note--object)\n    + [Getting Notes](#getting-notes)\n- [Contributing](#contributing)\n\n# Installation\n\n```python\npip install aiopinboard\n```\n\n# Python Versions\n\n`aiopinboard` is currently supported on:\n\n* Python 3.7\n* Python 3.8\n* Python 3.9\n\n# API Token\n\nYou can retrieve your Pinboard API token via\n[your account\'s settings page](https://pinboard.in/settings/password).\n\n# Usage\n\n`aiopinboard` endeavors to replicate all of the endpoints\n[the Pinboard API documentation](https://pinboard.in/api) with sane, usable responses.\n\nAll API usage starts with creating an `API` object that contains your Pinboard API token:\n\n```python\nimport asyncio\n\nfrom aiopinboard import Client\n\n\nasync def main() -> None:\n    api = API("<PINBOARD_API_TOKEN>")\n    # do things!\n\n\nasyncio.run(main())\n```\n\n## Bookmarks\n\n### The `Bookmark` Object\n\nAPI endpoints that retrieve one or more bookmarks will return `Bookmark` objects, which\ncarry all of the expected properties of a bookmark:\n\n* `hash`: the unique identifier of the bookmark\n* `href`: the bookmark\'s URL\n* `title`: the bookmark\'s title\n* `description`: the bookmark\'s description\n* `last_modified`: the UTC date the bookmark was last modified\n* `tags`: a list of tags applied to the bookmark\n* `unread`: whether the bookmark is unread\n* `shared`: whether the bookmark is shared\n\n### Getting the Last Change Datetime\n\nTo get the UTC datetime of the last "change" (bookmark added, updated, or deleted):\n\n```python\nimport asyncio\n\nfrom aiopinboard import Client\n\n\nasync def main() -> None:\n    api = API("<PINBOARD_API_TOKEN>")\n    last_change_dt = await.async_get_last_change_datetime()\n    # >>> datetime.datetime(2020, 9, 3, 13, 7, 19, tzinfo=<UTC>)\n\n\nasyncio.run(main())\n```\n\nThis method should be used to determine whether additional API calls should be made –\nfor example, if nothing has changed since the last time a request was made, the\nimplementing library can halt.\n\n### Getting Bookmarks\n\nTo get a bookmark by its URL:\n\n```python\nimport asyncio\n\nfrom aiopinboard import Client\n\n\nasync def main() -> None:\n    api = API("<PINBOARD_API_TOKEN>")\n    await api.bookmark.async_get_bookmark_by_url("https://my.com/bookmark")\n    # >>> <Bookmark href="https://my.com/bookmark">\n\n\nasyncio.run(main())\n```\n\nTo get all bookmarks\n\n```python\nimport asyncio\n\nfrom aiopinboard import Client\n\n\nasync def main() -> None:\n    api = API("<PINBOARD_API_TOKEN>")\n    await api.bookmark.async_get_all_bookmarks()\n    # >>> [<Bookmark ...>, <Bookmark ...>]\n\n\nasyncio.run(main())\n```\n\nYou can specify several optional parameters while editing a bookmark:\n\n* `tags`: an optional list of tags to filter results by\n* `start`: the optional starting index to return (defaults to the start)\n* `results`: the optional number of results (defaults to all)\n* `from_dt`: the optional datetime to start from\n* `to_dt`: the optional datetime to end at\n\nTo get all bookmarks created on a certain date:\n\n```python\nimport asyncio\nfrom datetime import date\n\nfrom aiopinboard import Client\n\n\nasync def main() -> None:\n    api = API("<PINBOARD_API_TOKEN>")\n    await api.bookmark.async_get_bookmarks_by_date(date.today())\n    # >>> [<Bookmark ...>, <Bookmark ...>]\n\n    # Optionally filter the results with a list of tags – note that only bookmarks that\n    # have all tags will be returned:\n    await api.bookmark.async_get_bookmarks_by_date(date.today(), tags=["tag1", "tag2"])\n    # >>> [<Bookmark ...>, <Bookmark ...>]\n)\n\n\nasyncio.run(main())\n```\n\nTo get recent bookmarks:\n\n```python\nimport asyncio\n\nfrom aiopinboard import Client\n\n\nasync def main() -> None:\n    api = API("<PINBOARD_API_TOKEN>")\n    await api.bookmark.async_get_recent_bookmarks(count=10)\n    # >>> [<Bookmark ...>, <Bookmark ...>]\n\n    # Optionally filter the results with a list of tags – note that only bookmarks that\n    # have all tags will be returned:\n    await api.bookmark.async_get_recent_bookmarks(count=20, tags=["tag1", "tag2"])\n    # >>> [<Bookmark ...>, <Bookmark ...>]\n\n\nasyncio.run(main())\n```\n\nTo get a summary of dates and how many bookmarks were created on those dates:\n\n```python\nimport asyncio\n\nfrom aiopinboard import Client\n\n\nasync def main() -> None:\n    api = API("<PINBOARD_API_TOKEN>")\n    dates = await api.bookmark.async_get_dates()\n    # >>> {datetime.date(2020, 09, 05): 4, ...}\n\n\nasyncio.run(main())\n```\n### Adding a Bookmark\n\nTo add a bookmark:\n\n```python\nimport asyncio\n\nfrom aiopinboard import Client\n\n\nasync def main() -> None:\n    api = API("<PINBOARD_API_TOKEN>")\n    await api.bookmark.async_add_bookmark("https://my.com/bookmark", "My New Bookmark")\n\n\nasyncio.run(main())\n```\n\nYou can specify several optional parameters while editing a bookmark:\n\n* `description`: the optional description of the bookmark\n* `tags`: an optional list of tags to assign to the bookmark\n* `created_datetime`: the optional creation datetime to use (defaults to now)\n* `replace`: whether this should replace a bookmark with the same URL\n* `shared`: whether this bookmark should be shared\n* `toread`: whether this bookmark should be unread\n\n### Deleting a Bookmark\n\nTo delete a bookmark by its URL:\n\n```python\nimport asyncio\n\nfrom aiopinboard import Client\n\n\nasync def main() -> None:\n    api = API("<PINBOARD_API_TOKEN>")\n    await api.bookmark.async_delete_bookmark("https://my.com/bookmark")\n\n\nasyncio.run(main())\n```\n\n## Tags\n\n### Getting Tags\n\nTo get all tags for an account (and a count of how often each tag is used):\n\n```python\nimport asyncio\n\nfrom aiopinboard import Client\n\n\nasync def main() -> None:\n    api = API("<PINBOARD_API_TOKEN>")\n    await api.tag.async_get_tags()\n    # >>> {"tag1": 3, "tag2": 8}\n\n\nasyncio.run(main())\n```\n\n### Getting Suggested Tags\n\nTo get lists of popular (used by the community) and recommended (used by you) tags for a\nparticular URL:\n\n```python\nimport asyncio\n\nfrom aiopinboard import Client\n\n\nasync def main() -> None:\n    api = API("<PINBOARD_API_TOKEN>")\n    await api.bookmark.async_get_suggested_tags("https://my.com/bookmark")\n    # >>> {"popular": ["tag1", "tag2"], "recommended": ["tag3"]}\n\n\nasyncio.run(main())\n```\n\n### Deleting a Tag\n\nTo delete a tag:\n\n```python\nimport asyncio\n\nfrom aiopinboard import Client\n\n\nasync def main() -> None:\n    api = API("<PINBOARD_API_TOKEN>")\n    await api.tag.async_delete_tag("tag1")\n\n\nasyncio.run(main())\n```\n\n### Renaming a Tag\n\nTo rename a tag:\n\n```python\nimport asyncio\n\nfrom aiopinboard import Client\n\n\nasync def main() -> None:\n    api = API("<PINBOARD_API_TOKEN>")\n    await api.tag.async_rename_tag("old-tag", "new-tag")\n\n\nasyncio.run(main())\n```\n\n## Notes\n\n### The `Note` Object\n\nAPI endpoints that retrieve one or more notes will return `Note` objects, which\ncarry all of the expected properties of a note:\n\n* `note_id`: the unique ID\n* `title`: the title\n* `hash`: the computed hash\n* `created_at`: the UTC datetime the note was created\n* `updated_at`: the UTC datetime the note was updated\n* `length`: the length\n\n### Getting Notes\n\nTo get all notes for an account:\n\n```python\nimport asyncio\n\nfrom aiopinboard import Client\n\n\nasync def main() -> None:\n    api = API("<PINBOARD_API_TOKEN>")\n    await api.note.async_get_notes()\n    # >>> [<Note ...>, <Note ...>]\n\n\nasyncio.run(main())\n```\n\n\n# Contributing\n\n1. [Check for open features/bugs](https://github.com/bachya/aiopinboard/issues)\n  or [initiate a discussion on one](https://github.com/bachya/aiopinboard/issues/new).\n2. [Fork the repository](https://github.com/bachya/aiopinboard/fork).\n3. (_optional, but highly recommended_) Create a virtual environment: `python3 -m venv .venv`\n4. (_optional, but highly recommended_) Enter the virtual environment: `source ./.venv/bin/activate`\n5. Install the dev environment: `script/setup`\n6. Code your new feature or bug fix.\n7. Write tests that cover your new functionality.\n8. Run tests and ensure 100% code coverage: `script/test`\n9. Update `README.md` with any new documentation.\n10. Add yourself to `AUTHORS.md`.\n11. Submit a pull request!\n',
    'author': 'Aaron Bach',
    'author_email': 'bachya1208@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bachya/aiopinboard',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)
